from scipy.optimize import linear_sum_assignment
import zebrazoom.dataAPI as dataAPI
import matplotlib.pyplot as plt
from scipy import stats
import pandas as pd
import numpy as np
import shutil
import math
import os


def identifyPointsToClose_old(curr_positions, threshold, dataForLine, numFrame):
  points = curr_positions[0]
  valid_points = points[~np.all(points == 0, axis=1)]
  distances = np.sqrt(np.sum((valid_points[:, np.newaxis, :] - valid_points[np.newaxis, :, :]) ** 2, axis=2))
  close_pairs = np.argwhere((distances < threshold) & (distances >= 0))
  close_pairs = [pair for pair in close_pairs if pair[0] < pair[1]]
  for _, second_idx in close_pairs:
    curr_positions[0, second_idx] = [0, 0]
    dataForLine[second_idx][numFrame] = [0, 0]

def identifyPointsToClose(curr_positions, threshold, dataForLine, numFrame):
  points = curr_positions[0]
  
  valid_indices = np.where(~np.all(points == 0, axis=1))[0]
  valid_points = points[valid_indices]
  
  distances = np.sqrt(np.sum((valid_points[:, np.newaxis, :] - valid_points[np.newaxis, :, :]) ** 2, axis=2))
  close_pairs = np.argwhere((distances < threshold) & (distances > 0))
  close_pairs = [pair for pair in close_pairs if pair[0] < pair[1]]
  
  points_to_remove = set()
  for _, second_idx in close_pairs:
    global_second_idx = valid_indices[second_idx]
    points_to_remove.add(global_second_idx)
  
  for idx in points_to_remove:
    curr_positions[0, idx] = [0, 0]
    dataForLine[idx][numFrame] = [0, 0]


def reassignMultipleAnimalsId(videoName: str, nbWells: int, nbAnimalsPerWell: int, freqAlgoPosFollow: int, startTimeInSeconds = None, endTimeInSeconds = None, max_distance_threshold = 100, max_dist_disapearedAnimal_step = 0, max_NbFramesAllowedToDisapeared = 50, minimumTraceLength = 5, removeNewDetectionsTooClose = 8, minDistTravel = 1, minimumProbaDetectionForNewTrajectory = 0, removeStationaryPointMinDist = 0, removeStationaryPointInterval = 0):
  
  # NEED TO REMOVE THE FOLLOWING LINE AS SOON AS GETDATAPERTIMEINTERVAL AND SETDATAPERTIMEINTERVAL ARE FIXED
  dataAPI.setFPSandPixelSize(videoName, 24, 0.1)

  nbWells   = 1

  printDebug = False

  for numWell in range(nbWells):
    
    dataForEachAnimal = []
    probaForEachAnimal = []
    for numAnimal in range(nbAnimalsPerWell):
      data = dataAPI.getDataPerTimeInterval(videoName, numWell, numAnimal, startTimeInSeconds, endTimeInSeconds, "HeadPos")
      dataForEachAnimal.append(data)
      if minimumProbaDetectionForNewTrajectory:
        proba = dataAPI.getDataPerTimeInterval(videoName, numWell, numAnimal, startTimeInSeconds, endTimeInSeconds, "probabilityGoodDetection")
        probaForEachAnimal.append(proba)
    
    maximalDisappearanceFrames = [2 * max_NbFramesAllowedToDisapeared for i in range(nbAnimalsPerWell)]
    nbFramesNotAtZero = [0 for i in range(nbAnimalsPerWell)]
    curFirstCoordinates = [[0, 0] for i in range(nbAnimalsPerWell)]
    
    # No new traces allowed for proba too low
    numFrame = 0
    if minimumProbaDetectionForNewTrajectory > 0:
      for numAnimal in range(nbAnimalsPerWell):
        if probaForEachAnimal[numAnimal][numFrame] < minimumProbaDetectionForNewTrajectory:
          dataForEachAnimal[numAnimal][numFrame] = [0, 0]
    
    for numFrame in range(1, len(dataForEachAnimal[0])):
      
      if numFrame % freqAlgoPosFollow == 0:
        print("reassignMultipleIds: numFrame:", numFrame)
      
      prev_positions = np.empty((0, 1, 2))
      curr_positions = np.empty((0, 2))
      for numAnimal in range(nbAnimalsPerWell):
        prev_positions = np.append(prev_positions, np.array([np.array([dataForEachAnimal[numAnimal][numFrame-1]])]), axis = 0)
        curr_positions = np.append(curr_positions, np.array([dataForEachAnimal[numAnimal][numFrame]]), axis = 0)
      curr_positions = np.array([curr_positions])
      if removeNewDetectionsTooClose > 0:
        identifyPointsToClose(curr_positions, removeNewDetectionsTooClose, dataForEachAnimal, numFrame)
      
      distance_matrix = np.linalg.norm(prev_positions - curr_positions, axis=2)
      
      # Identifying next possible positions equal to (0, 0) or not
      non_zero_mask = np.linalg.norm(curr_positions, axis=2) != 0
      zero_mask = np.linalg.norm(curr_positions, axis=2) == 0
      
      # Adding disavantage for long distance position changes 
      # but only for next possible positions not at (0, 0) (because we want it to be possible for the trace to "end")
      # typically this will results in the first columns (possible next coordinates) being penalized for distances too long (first lines because traces currently at (0, 0), last lines because they could be simply too far)
      if max_dist_disapearedAnimal_step > 0:
        max_distance_threshold_array = np.zeros((len(distance_matrix), len(distance_matrix[0])))
        for numAnimal in range(len(max_distance_threshold_array)):
          max_distance_threshold_array[numAnimal, :] = max_distance_threshold + max_dist_disapearedAnimal_step * maximalDisappearanceFrames[numAnimal]
        distance_matrix[(distance_matrix > max_distance_threshold_array) & non_zero_mask] += 1e6
      else:
        distance_matrix[(distance_matrix > max_distance_threshold) & non_zero_mask] += 1e6
      
      # Making all end of trace "return to (0, 0)" equal regardless of current position
      distance_matrix[(distance_matrix != 0) & zero_mask] = 1e4 #0.5 * distance_matrix[(distance_matrix != 0) & zero_mask]
      
      # Adding disavantage for traces at (0, 0) to stay at (0, 0) (to promote traces to "start")
      # typically this will result in the first lines (not yet assigned to a trace) being penalized for the last columns (which are (0, 0) next coordinates options) (starting from numcolumn = number of possible next coordinates options)
      distance_matrix[(distance_matrix == 0) & zero_mask] += 1e6
      
      row_ind, col_ind = linear_sum_assignment(distance_matrix)
      
      if printDebug:
        print("row_ind, col_ind:", row_ind, col_ind)
        print(numFrame, [dataForEachAnimal[idAnimal][numFrame] for idAnimal in range(nbAnimalsPerWell)])
      
      if np.sum(row_ind == col_ind) != nbAnimalsPerWell:
        newPossibilities = [dataForEachAnimal[i][numFrame].copy() for i in range(len(dataForEachAnimal))]
        if minimumProbaDetectionForNewTrajectory:
          shiftedProba     = [probaForEachAnimal[i][numFrame].copy() for i in range(len(probaForEachAnimal))]
        for idx, rowId in enumerate(row_ind):
          colId = col_ind[idx]
          if rowId != colId:
            dataForEachAnimal[rowId][numFrame]  = newPossibilities[colId]
            if minimumProbaDetectionForNewTrajectory:
              probaForEachAnimal[rowId][numFrame] = shiftedProba[colId]
      
      if printDebug:
        print(numFrame, [dataForEachAnimal[idAnimal][numFrame] for idAnimal in range(nbAnimalsPerWell)])
      
      for numAnimal in range(nbAnimalsPerWell):
        if np.sum(dataForEachAnimal[numAnimal][numFrame - 1] == [0, 0]) != 2 and np.sum(dataForEachAnimal[numAnimal][numFrame] == [0, 0]) == 2: # No candidate point was found for the current numAnimal trace
          if maximalDisappearanceFrames[numAnimal] < max_NbFramesAllowedToDisapeared:
            dataForEachAnimal[numAnimal][numFrame] = dataForEachAnimal[numAnimal][numFrame - 1]
            maximalDisappearanceFrames[numAnimal] += 1
          else: # Trace was lost after max_NbFramesAllowedToDisapeared disappearance: removing points artificially added
            nbFrameDisappeared = maximalDisappearanceFrames[numAnimal]
            
            distTravelMat = dataForEachAnimal[numAnimal][max(0, numFrame - nbFrameDisappeared - nbFramesNotAtZero[numAnimal] + 2)] - dataForEachAnimal[numAnimal][max(0, numFrame - nbFrameDisappeared)]
            # distTravelMat = curFirstCoordinates[numAnimal] - dataForEachAnimal[numAnimal][max(0, numFrame - nbFrameDisappeared)]
            distTravel    = np.sqrt(np.sum(distTravelMat**2))
            
            nbRemoveMoreFramesAsTraceTooShort = 0 if (nbFramesNotAtZero[numAnimal] >= minimumTraceLength and distTravel > minDistTravel) else nbFramesNotAtZero[numAnimal]
            
            for i in range(max(0, numFrame - nbFrameDisappeared - nbRemoveMoreFramesAsTraceTooShort - 1), numFrame + 1):
              dataForEachAnimal[numAnimal][i][0] = 0
              dataForEachAnimal[numAnimal][i][1] = 0
            maximalDisappearanceFrames[numAnimal] = 0
            nbFramesNotAtZero[numAnimal] = 0
        else:
          if maximalDisappearanceFrames[numAnimal] != 0 and maximalDisappearanceFrames[numAnimal] != 2 * max_NbFramesAllowedToDisapeared: # A point had disappeared: rebuilding the lost points
            nbFrameDisappeared = maximalDisappearanceFrames[numAnimal]
            xOri = dataForEachAnimal[numAnimal][numFrame - nbFrameDisappeared][0]
            yOri = dataForEachAnimal[numAnimal][numFrame - nbFrameDisappeared][1]
            xEnd = dataForEachAnimal[numAnimal][numFrame][0]
            yEnd = dataForEachAnimal[numAnimal][numFrame][1]
            for i in range(numFrame - nbFrameDisappeared, numFrame):
              dataForEachAnimal[numAnimal][i][0] = xOri + ((i - (numFrame - nbFrameDisappeared)) / nbFrameDisappeared) * (xEnd - xOri)
              dataForEachAnimal[numAnimal][i][1] = yOri + ((i - (numFrame - nbFrameDisappeared)) / nbFrameDisappeared) * (yEnd - yOri)
            nbFramesNotAtZero[numAnimal] += nbFrameDisappeared
          else:
            if np.sum(dataForEachAnimal[numAnimal][numFrame] == [0, 0]) != 2:
              nbFramesNotAtZero[numAnimal] += 1
          maximalDisappearanceFrames[numAnimal] = 0
      
      # No new traces allowed for proba too low
      if minimumProbaDetectionForNewTrajectory > 0:
        for numAnimal in range(nbAnimalsPerWell):
          if (numFrame == 1 or np.sum(dataForEachAnimal[numAnimal][numFrame - 1] == [0, 0]) == 2) and probaForEachAnimal[numAnimal][numFrame] < minimumProbaDetectionForNewTrajectory:
            dataForEachAnimal[numAnimal][numFrame] = [0, 0]
      
      if printDebug:
        print(numFrame, [dataForEachAnimal[idAnimal][numFrame] for idAnimal in range(nbAnimalsPerWell)])
    
    if removeStationaryPointMinDist and removeStationaryPointInterval:
    
      for numFrame in range(1, len(dataForEachAnimal[0])):
        
        if numFrame % freqAlgoPosFollow == 0:
          print("reassignMultipleIds: numFrame:", numFrame)
          
        # Removing objects not moving enough
        if removeStationaryPointMinDist and numFrame > removeStationaryPointInterval * 2 and numFrame % removeStationaryPointInterval == 0:
          for numAnimal in range(nbAnimalsPerWell):
            if not(dataForEachAnimal[numAnimal][numFrame][0] == 0 and dataForEachAnimal[numAnimal][numFrame][1] == 0):
              distTravelMat1 = dataForEachAnimal[numAnimal][numFrame-removeStationaryPointInterval] - dataForEachAnimal[numAnimal][numFrame]
              distTravel1    = np.sqrt(np.sum(distTravelMat1**2))
              if distTravel1 < removeStationaryPointMinDist:
                distTravelMat2 = dataForEachAnimal[numAnimal][numFrame - 2 * removeStationaryPointInterval] - dataForEachAnimal[numAnimal][numFrame]
                distTravel2    = np.sqrt(np.sum(distTravelMat2**2))
                if distTravel2 < removeStationaryPointMinDist:
                  distTravel = 0
                  startDelete = numFrame - 2 * removeStationaryPointInterval
                  while startDelete >= 1 and distTravel < removeStationaryPointMinDist:
                    distTravelMat = dataForEachAnimal[numAnimal][startDelete] - dataForEachAnimal[numAnimal][numFrame]
                    distTravel    = np.sqrt(np.sum(distTravelMat**2))
                    startDelete -= 1
                  distTravel = 0
                  stopDelete = numFrame
                  while stopDelete < len(dataForEachAnimal[numAnimal]) and distTravel < removeStationaryPointMinDist:
                    distTravelMat = dataForEachAnimal[numAnimal][stopDelete] - dataForEachAnimal[numAnimal][numFrame]
                    distTravel    = np.sqrt(np.sum(distTravelMat**2))
                    stopDelete += 1
                  for i in range(startDelete, stopDelete):
                    dataForEachAnimal[numAnimal][i][0] = 0
                    dataForEachAnimal[numAnimal][i][1] = 0
    
    for numAnimal in range(nbAnimalsPerWell):
      if maximalDisappearanceFrames[numAnimal] != 0: # Removing artifically added points if the trace had not been "saved" by the end of the video
        nbFrameDisappeared = maximalDisappearanceFrames[numAnimal]
        for i in range(numFrame - nbFrameDisappeared, numFrame):
          dataForEachAnimal[numAnimal][i][0] = 0
          dataForEachAnimal[numAnimal][i][1] = 0
    
    for numAnimal in range(nbAnimalsPerWell):
      data = dataForEachAnimal[numAnimal]
      dataAPI.setDataPerTimeInterval(videoName, numWell, numAnimal, startTimeInSeconds, endTimeInSeconds, "HeadPos", data)



from zebrazoom.code.tracking.customTrackingImplementations.yolov11.calculateHeading import calculateHeading
import numpy as np
import cv2

def invertSkeletonIfNecessaryUsingTheDarkEyes(image, currContour, skeleton_points):

  if len(skeleton_points) == 0:
    return skeleton_points

  mask = np.zeros_like(image, dtype=np.uint8)
  currContour = currContour.astype(np.int32)
  cv2.fillPoly(mask, [currContour], 255)
  ys, xs = np.where(mask == 255)
  pixel_values = image[ys, xs]
  sorted_indices = np.argsort(pixel_values)
  darkest_points = list(zip(xs[sorted_indices[:10]], ys[sorted_indices[:10]]))
  darkest_points_array = np.array(darkest_points)
  mean_x = np.mean(darkest_points_array[:, 0])
  mean_y = np.mean(darkest_points_array[:, 1])
  d1 = np.sum((skeleton_points[:, 0, :][0] - np.array([mean_x, mean_y]))**2)
  dN = np.sum((skeleton_points[:, 0, :][len(skeleton_points)-1] - np.array([mean_x, mean_y]))**2)
  if dN < d1:
    mirror_points = np.copy(skeleton_points)
    mirror_points[:, 0] = skeleton_points[:, 0][::-1]
    skeleton_points = mirror_points
  return skeleton_points


def invertSkeletonIfNecessaryUsingThePast(self, skeleton_points_cur, wellNum, frameNum, numAlreadyInvertedWithThePast, idxContour, maxConsecutiveFramesInversion=20):
  
  if len(skeleton_points_cur) == 0:
    numAlreadyInvertedWithThePast[idxContour] = 0
    return skeleton_points_cur
  
  skeleton_points_past = self._trackingHeadTailAllAnimalsList[wellNum][idxContour][frameNum-self._firstFrame-1]
  
  if numAlreadyInvertedWithThePast[idxContour] < maxConsecutiveFramesInversion: # Cannot invert using the past for more than maxConsecutiveFramesInversion consecutive frames
    mirror_points = np.copy(skeleton_points_cur)
    mirror_points[:, 0] = skeleton_points_cur[:, 0][::-1]
    
    total_distance = np.sum(np.linalg.norm(skeleton_points_cur.reshape(-1, 2) - skeleton_points_past, axis=1))
    total_distance_mirror = np.sum(np.linalg.norm(mirror_points.reshape(-1, 2) - skeleton_points_past, axis=1))
    
    if total_distance < total_distance_mirror:
      numAlreadyInvertedWithThePast[idxContour] = 0
      return skeleton_points_cur
    else:
      numAlreadyInvertedWithThePast[idxContour] += 1
      return mirror_points
  
  else:
    
    # Reverting back to original if maxConsecutiveFramesInversion is reached
    backIdx = frameNum-self._firstFrame-1
    while backIdx > max(0, frameNum-self._firstFrame-maxConsecutiveFramesInversion):
      skeleton_points_past = self._trackingHeadTailAllAnimalsList[wellNum][idxContour][backIdx]
      skeleton_points_past_mirror = np.flip(skeleton_points_past, 0)
      self._trackingHeadTailAllAnimalsList[wellNum][idxContour][backIdx] = skeleton_points_past_mirror
      self._trackingHeadingAllAnimalsList[wellNum][idxContour][backIdx] = calculateHeading(self._trackingHeadTailAllAnimalsList[wellNum][idxContour][backIdx])
      backIdx -= 1
    
    numAlreadyInvertedWithThePast[idxContour] = 0
    return skeleton_points_cur



import math
import cv2
import numpy as np


def _headTrackingHeadingCalculation(self, i, blur, thresh1, thresh2, frameOri, erodeSize, frame_width, frame_height, trackingHeadingAllAnimals, trackingHeadTailAllAnimals, trackingProbabilityOfGoodDetection, headPosition, lengthX, xmin=0, ymin=0, wellNumber=-1, oldFrameList=[]):
  xHB_TN = 0
  heading = 0
  x = 0
  y = 0
  lastFirstTheta = 0

  # Finds head position for frame i
  takeTheHeadClosestToTheCenter = self._hyperparameters["takeTheHeadClosestToTheCenter"]
  
  (minVal, maxVal, headPosition, maxLoc) = cv2.minMaxLoc(blur)
  
  if "localMinimumDarkestThreshold" in self._hyperparameters and self._hyperparameters["localMinimumDarkestThreshold"]:
    localMinimumDarkestThreshold = int(self._hyperparameters["localMinimumDarkestThreshold"])
  else:
    localMinimumDarkestThreshold = 180
  
  for animalNumber in range(self._hyperparameters["nbAnimalsPerWell"]):
    
    if minVal < localMinimumDarkestThreshold:
      
      if type(trackingProbabilityOfGoodDetection) != int and len(trackingProbabilityOfGoodDetection) and i-self._firstFrame < len(trackingProbabilityOfGoodDetection[0]):
        trackingProbabilityOfGoodDetection[0, i-self._firstFrame] = np.sum(255 - blur)

      x = headPosition[0]
      y = headPosition[1]

      if type(headPosition) == tuple:
        headPosition = list(headPosition)
        headPosition[0] = headPosition[0] + xmin
        headPosition[1] = headPosition[1] + ymin
        headPosition = tuple(headPosition)
      else:
        headPosition[0] = headPosition[0] + xmin
        headPosition[1] = headPosition[1] + ymin

      # Calculate heading for frame i
      if type(thresh1) != int:
        [heading, lastFirstTheta] = self._calculateHeading(x, y, i, thresh1, thresh2, takeTheHeadClosestToTheCenter, 0, wellNumber)
      
      okAll = True
      if "removeShades" in self._hyperparameters and self._hyperparameters["removeShades"]:
        if i > 110:
          for oldFrame in oldFrameList:
            compareDarknestWithThePastWindow = int(self._hyperparameters["headSize"] / 2) if int(self._hyperparameters["headSize"] / 2) >= 2 else 2
            centeredROIValue = np.mean(frameOri[headPosition[1]-compareDarknestWithThePastWindow:headPosition[1]+compareDarknestWithThePastWindow, headPosition[0]-compareDarknestWithThePastWindow:headPosition[0]+compareDarknestWithThePastWindow])
            centeredROIValueOld = np.mean(oldFrame[headPosition[1]-compareDarknestWithThePastWindow:headPosition[1]+compareDarknestWithThePastWindow, headPosition[0]-compareDarknestWithThePastWindow:headPosition[0]+compareDarknestWithThePastWindow])
            if self._hyperparameters["brightAnimalDarkBackground"]:
              ok = (centeredROIValue > centeredROIValueOld)
            else:
              ok = (centeredROIValue < centeredROIValueOld)
            okAll = (okAll and ok)
      
      if minVal < localMinimumDarkestThreshold and okAll:
        trackingHeadTailAllAnimals[animalNumber, i-self._firstFrame][0][0] = headPosition[0]
        trackingHeadTailAllAnimals[animalNumber, i-self._firstFrame][0][1] = headPosition[1]
        trackingHeadingAllAnimals[animalNumber, i-self._firstFrame] = heading
      else:
        trackingHeadTailAllAnimals[animalNumber, i-self._firstFrame][0][0] = 0
        trackingHeadTailAllAnimals[animalNumber, i-self._firstFrame][0][1] = 0
        trackingHeadingAllAnimals[animalNumber, i-self._firstFrame] = 0
      
      cv2.circle(blur, headPosition, self._hyperparameters["headSize"], (255, 255, 255), -1)
      (minVal, maxVal, headPosition, maxLoc) = cv2.minMaxLoc(blur)

  return lastFirstTheta



from zebrazoom.code.extractParameters import extractParameters

import cv2
import zebrazoom.videoFormatConversion.zzVideoReading as zzVideoReading
import numpy as np
import math
from zebrazoom.code.tracking import register_tracking_method
from ._eyeTracking import EyeTrackingMixin
from ._fasterMultiprocessingBase import BaseFasterMultiprocessing
from ._getImages import GetImagesMixin

import zebrazoom.code.util as util


class FasterMultiprocessing(BaseFasterMultiprocessing, EyeTrackingMixin, GetImagesMixin):
  def __init__(self, videoPath, wellPositions, hyperparameters):
    super().__init__(videoPath, wellPositions, hyperparameters)
    if self._hyperparameters["eyeTracking"]:
      self._trackingEyesAllAnimalsList = [np.zeros((self._hyperparameters["nbAnimalsPerWell"], self._lastFrame-self._firstFrame+1, 8))
                                          for _ in range(self._hyperparameters["nbWells"])]
    else:
      self._trackingEyesAllAnimals = 0

    if not(self._hyperparameters["nbAnimalsPerWell"] > 1) and not(self._hyperparameters["headEmbeded"]) and (self._hyperparameters["findHeadPositionByUserInput"] == 0) and (self._hyperparameters["takeTheHeadClosestToTheCenter"] == 0):
      self._trackingProbabilityOfGoodDetectionList = [np.zeros((self._hyperparameters["nbAnimalsPerWell"], self._lastFrame-self._firstFrame+1))
                                                      for _ in range(self._hyperparameters["nbWells"])]
    else:
      self._trackingProbabilityOfGoodDetectionList = 0
    
    if self._hyperparameters["headingCalculationMethod"]:
      self._trackingProbabilityOfHeadingGoodCalculation = [np.zeros((self._hyperparameters["nbAnimalsPerWell"], self._lastFrame-self._firstFrame+1)) for _ in range(self._hyperparameters["nbWells"])]
    else:
      self._trackingProbabilityOfHeadingGoodCalculation = 0

  def _adjustParameters(self, i, frame, widgets):
    return None

  def _formatOutput(self):
    if self._hyperparameters["postProcessMultipleTrajectories"]:
      for wellNumber in range(self._firstWell, self._lastWell + 1):
        self._postProcessMultipleTrajectories(self._trackingHeadTailAllAnimalsList[wellNumber], self._trackingProbabilityOfGoodDetectionList[wellNumber])
    
    if "postProcessHeadingWithTrajectoryAdvanced" in self._hyperparameters and self._hyperparameters["postProcessHeadingWithTrajectoryAdvanced"]:
      for wellNumber in range(self._firstWell, self._lastWell + 1):
        self._postProcessHeadingWithTrajectoryAdvanced(self._trackingHeadTailAllAnimalsList[wellNumber], self._trackingHeadingAllAnimalsList[wellNumber], self._trackingProbabilityOfHeadingGoodCalculation[wellNumber])
    
    if "postProcessHeadingWithTrajectory_minDist" in self._hyperparameters and self._hyperparameters["postProcessHeadingWithTrajectory_minDist"]:
      for wellNumber in range(self._firstWell, self._lastWell + 1):
        self._postProcessHeadingWithTrajectory(self._trackingHeadTailAllAnimalsList[wellNumber], self._trackingHeadingAllAnimalsList[wellNumber], self._trackingProbabilityOfHeadingGoodCalculation[wellNumber])
    
    return {wellNumber: extractParameters([self._trackingHeadTailAllAnimalsList[wellNumber], self._trackingHeadingAllAnimalsList[wellNumber], [], 0, 0] + ([self._auDessusPerAnimalIdList[wellNumber]] if self._auDessusPerAnimalIdList is not None else []), wellNumber, self._hyperparameters, self._videoPath, self._wellPositions, self._background)
            for wellNumber in range(self._firstWell, self._lastWell + 1)}
  
  def _setToWhiteAllPixelsTooFarFromTheCenter(self, image, maxDist, wellNumber):
    
    if hasattr(self, 'maskForTooFarFromCenter') and wellNumber < len(self.maskForTooFarFromCenter):
      mask = self.maskForTooFarFromCenter[wellNumber]
    else:
      height, width = image.shape[:2]
      center_x, center_y = width // 2, height // 2
      y, x = np.ogrid[:height, :width]
      distance_from_center = np.sqrt((x - center_x) ** 2 + (y - center_y) ** 2)
      mask = distance_from_center > maxDist
      if not(hasattr(self, 'maskForTooFarFromCenter')):
        self.maskForTooFarFromCenter = []
      self.maskForTooFarFromCenter.append(mask)
    
    image[mask] = 255
    return image
  
  def run(self):
    self._background = self.getBackground()

    cap = zzVideoReading.VideoCapture(self._videoPath)
    if (cap.isOpened()== False):
      print("Error opening video stream or file")

    if self._hyperparameters["backgroundSubtractorKNN"]:
      fgbg = cv2.createBackgroundSubtractorKNN()
      for i in range(0, min(self._lastFrame - 1, 500), int(min(self._lastFrame - 1, 500) / 10)):
        cap.set(1, min(self._lastFrame - 1, 500) - i)
        ret, frame = cap.read()
        fgmask = fgbg.apply(frame)
      cap.release()
      cap = zzVideoReading.VideoCapture(self._videoPath)

    i = self._firstFrame

    if self._firstFrame:
      cap.set(1, self._firstFrame)

    previousFrames = None
    widgets = None
    while (i < self._lastFrame + 1):

      if (self._hyperparameters["freqAlgoPosFollow"] != 0) and (i % self._hyperparameters["freqAlgoPosFollow"] == 0):
        print("Tracking: frame:",i)
        if self._hyperparameters["popUpAlgoFollow"]:
          from zebrazoom.code.popUpAlgoFollow import prepend

          prepend("Tracking: frame:" + str(i))

      if self._hyperparameters["debugTracking"]:
        print("frame:",i)

      ret, frame = cap.read()

      if ret:
        
        if self._hyperparameters.get("useOriginalImageForHeadingCalculation", False) or self._hyperparameters["detectMovementWithRawVideoInsideTracking"]:
          frameOri = frame.copy()
          frameOri = cv2.cvtColor(frameOri, cv2.COLOR_BGR2GRAY)
        
        if self._hyperparameters["backgroundSubtractorKNN"]:
          frame = fgbg.apply(frame)
          frame = 255 - frame

        for wellNumber in range(self._firstWell, self._lastWell + 1):
          if self._hyperparameters["nbAnimalsPerWell"] == 1 and not(self._hyperparameters["forceBlobMethodForHeadTracking"]):
            minPixelDiffForBackExtract = self._hyperparameters["minPixelDiffForBackExtract"]
            xtop = self._wellPositions[wellNumber]['topLeftX']
            ytop = self._wellPositions[wellNumber]['topLeftY']
            lenX = self._wellPositions[wellNumber]['lengthX']
            lenY = self._wellPositions[wellNumber]['lengthY']
            if self._hyperparameters["backgroundSubtractorKNN"]:
              grey = frame
            else:
              grey = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)

            curFrame = grey[ytop:ytop+lenY, xtop:xtop+lenX].copy()
            if "setToWhiteAllPixelsTooFarFromTheCenter" in self._hyperparameters and self._hyperparameters["setToWhiteAllPixelsTooFarFromTheCenter"]:
              curFrame = self._setToWhiteAllPixelsTooFarFromTheCenter(curFrame, self._hyperparameters["setToWhiteAllPixelsTooFarFromTheCenter"] if not(self._hyperparameters.get("setToWhiteAllPixelsTooFarFromTheCenter_takeROIsWidth", 0)) else int(lenX/2), wellNumber)
            
            if not(self._hyperparameters["backgroundSubtractorKNN"]):
              back = self._background[ytop:ytop+lenY, xtop:xtop+lenX]
              putToWhite = ( curFrame.astype('int32') >= (back.astype('int32') - minPixelDiffForBackExtract) )
              curFrame[putToWhite] = 255
            else:
              self._hyperparameters["paramGaussianBlur"] = int(math.sqrt(cv2.countNonZero(255 - curFrame) / self._hyperparameters["nbAnimalsPerWell"]) / 2) * 2 + 1
            if self._hyperparameters["paramGaussianBlur"]:
              blur = cv2.GaussianBlur(curFrame, (self._hyperparameters["paramGaussianBlur"], self._hyperparameters["paramGaussianBlur"]),0)
            else:
              blur = curFrame
            if "headingCalculationMethod" in self._hyperparameters:
            
              if "useOriginalImageForHeadingCalculation" in self._hyperparameters and self._hyperparameters["useOriginalImageForHeadingCalculation"]:
                grey2 = frameOri.copy()
                curFrame2 = grey2[ytop:ytop+lenY, xtop:xtop+lenX]
                back = self._background[ytop:ytop+lenY, xtop:xtop+lenX]
                if self._hyperparameters["debugExtractBack"]:
                  util.showFrame(self._setToWhiteAllPixelsTooFarFromTheCenter(back, self._hyperparameters["setToWhiteAllPixelsTooFarFromTheCenter"], wellNumber), title="back")
                putToWhite = ( curFrame2.astype('int32') >= (back.astype('int32') - minPixelDiffForBackExtract) )
                curFrame2[putToWhite] = 255
                if "setToWhiteAllPixelsTooFarFromTheCenter" in self._hyperparameters and self._hyperparameters["setToWhiteAllPixelsTooFarFromTheCenter"]:
                  curFrame2 = self._setToWhiteAllPixelsTooFarFromTheCenter(curFrame2, self._hyperparameters["setToWhiteAllPixelsTooFarFromTheCenter"], wellNumber)
                t, thresh1 = cv2.threshold(curFrame2, 254, 255, cv2.THRESH_BINARY)
                t, thresh2 = cv2.threshold(curFrame2, 254, 255, cv2.THRESH_BINARY)
              else:
                t, thresh1 = cv2.threshold(curFrame, 254, 255, cv2.THRESH_BINARY)
                t, thresh2 = cv2.threshold(curFrame, 254, 255, cv2.THRESH_BINARY)
              
            else:
              thresh1 = 0
              thresh2 = 0
            gray    = 0
          else:
            [frame2, gray, thresh1, blur, thresh2, frame2, initialCurFrame, back, xHead, yHead] = self._getImages(0, i, wellNumber, frame)

          headPositionFirstFrame = 0

          # Head tracking and heading calculation
          lastFirstTheta = self._headTrackingHeadingCalculation(i, blur, thresh1, thresh2, gray, self._hyperparameters["erodeSize"], int(cap.get(3)), int(cap.get(4)), self._trackingHeadingAllAnimalsList[wellNumber], self._trackingHeadTailAllAnimalsList[wellNumber], self._trackingProbabilityOfGoodDetectionList[wellNumber], headPositionFirstFrame, self._wellPositions[wellNumber]["lengthX"], 0, 0, wellNumber)

          # Tail tracking for frame i
          if self._hyperparameters["trackTail"] == 1 :
            threshForBlackFrames = 0
            thetaDiffAccept = 1.2
            lastFirstTheta = 0
            maxDepth = 0
            tailTipFirstFrame = []
            for animalId in range(0, self._hyperparameters["nbAnimalsPerWell"]):
              self._tailTracking(animalId, i, frame, thresh1, threshForBlackFrames, thetaDiffAccept, self._trackingHeadTailAllAnimalsList[wellNumber], self._trackingHeadingAllAnimalsList[wellNumber], lastFirstTheta, maxDepth, tailTipFirstFrame, initialCurFrame.copy(), back)

          # Eye tracking for frame i
          if self._hyperparameters["eyeTracking"]:
            self._eyeTracking(animalId, i, frame, thresh1, self._trackingHeadingAllAnimalsList[wellNumber], self._trackingHeadTailAllAnimalsList[wellNumber], self._trackingHeadingAllAnimalsList[wellNumber])

          self._debugTracking(i, self._trackingHeadTailAllAnimalsList[wellNumber], self._trackingHeadingAllAnimalsList[wellNumber], curFrame)

          if self._hyperparameters["freqAlgoPosFollow"]:
            if i % self._hyperparameters["freqAlgoPosFollow"] == 0:
              print("Tracking at frame", i)

        if self._hyperparameters["detectMovementWithRawVideoInsideTracking"]:
          previousFrames = self._detectMovementWithRawVideoInsideTracking(i, frameOri, previousFrames)
      
      paramsAdjusted = self._adjustParameters(i, frame, widgets)
      if paramsAdjusted is not None:
        i, widgets = paramsAdjusted
        cap.set(1, i)
      else:
        i = i + 1
    
    if self._hyperparameters["detectMovementWithRawVideoInsideTracking"]:
      frameGapComparision = self._hyperparameters["frameGapComparision"]
      nbFrames = len(self._trackingHeadTailAllAnimalsList[0][0])
      for wellNumber in range(len(self._trackingHeadTailAllAnimalsList)):
        for animalId in range(len(self._trackingHeadTailAllAnimalsList[wellNumber])):
          self._auDessusPerAnimalIdList[wellNumber][animalId][:nbFrames-frameGapComparision] = self._auDessusPerAnimalIdList[wellNumber][animalId][frameGapComparision:nbFrames]
      
    cap.release()
    return self._formatOutput()


register_tracking_method('fasterMultiprocessing', FasterMultiprocessing)



import cv2

import zebrazoom.code.tracking
from zebrazoom.code.adjustHyperparameters import adjustHyperparameters


baseClass = zebrazoom.code.tracking.get_tracking_method('multipleCenterOfMassTracking')


class GUITracking(baseClass):
  def _adjustParameters(self, i, frame, unprocessedFrame, widgets):
    if not self._hyperparameters['adjustFreelySwimTracking']:
      return None

    hyperparametersListNames = ["Red dot size", "backgroundSubtractorKNN_history", "backgroundSubtractorKNN_dist2Threshold", "paramGaussianBlur", "localMinimumDarkestThreshold", "headSize"]
    organizationTab = [
    [1, 10, "Size of the red dot in pixels."],
    [1, 30, "Number of previous frames taken into account to construct background model."],
    [1, 400, "Threshold for foreground pixel detection: increase to detect less foreground pixels."],
    [1, 50, "Window of gaussian filter applied on the image."],
    [1, 400, "Increase to detect smaller animals, decrease to avoid detecting them."],
    [1, self._hyperparameters['headSize'] * 1.3, "Minimum radius of animal, increase if multiple animals are detected where only one is present."],]
    title = "Adjust parameters in order for the background to be white and the animals to be gray/black."

    frame2 = unprocessedFrame if widgets is not None and 'unprocessedFrameCheckbox' in widgets and widgets['unprocessedFrameCheckbox'].isChecked() else cv2.cvtColor(frame, cv2.COLOR_GRAY2RGB)

    zoomInCheckbox = None if widgets is None or 'zoomInCheckbox' not in widgets else widgets['zoomInCheckbox']
    zoomInCoordinates = None
    for wellIdx in range(self._firstWell, self._lastWell + 1):
      for k in range(0, self._hyperparameters["nbAnimalsPerWell"]):
        output = self._trackingHeadTailAllAnimalsList[wellIdx][k]
        x = int(output[i-self._firstFrame][0][0])
        y = int(output[i-self._firstFrame][0][1])
        cv2.circle(frame2, (x, y), 2 if widgets is None else widgets['Red dot size'].value(), (0, 0, 255),   -1)
        if zoomInCheckbox is not None and zoomInCheckbox.isChecked() and zoomInCheckbox.getAnimalIdx() == k:
          zoomInCoordinates = (x, y)

    if zoomInCoordinates is not None:
      x, y = zoomInCoordinates
      halfLength = 125.5
      xmin = int(x - halfLength)
      xmax = int(x + halfLength)
      ymin = int(y - halfLength)
      ymax = int(y + halfLength)

      x = max(xmin, 0)
      y = max(ymin, 0)
      lengthX = xmax - xmin
      lengthY = ymax - ymin

      frameHeight, frameWidth = frame.shape[:2]
      if y + lengthY >= frameHeight:
        lengthY = frameHeight - y - 1
      if x + lengthX >= frameWidth:
        lengthX = frameWidth - x - 1

      frame2 = frame2[y:y+lengthY, x:x+lengthX]

    self._hyperparameters['Red dot size'] = 2 if widgets is None else widgets['Red dot size'].value()
    i, widgets = adjustHyperparameters(i, self._hyperparameters, hyperparametersListNames, frame2, title, organizationTab, widgets, addZoomCheckbox=True, addUnprocessedFrameCheckbox=True)
    del self._hyperparameters['Red dot size']
    return i, widgets

zebrazoom.code.tracking.register_tracking_method('multipleCenterOfMassTracking', GUITracking)


from skimage.morphology import skeletonize
import numpy as np
import cv2


def farthest_point(endpoints):
  max_distance = 0
  farthest_pt = None
  
  for i, p in enumerate(endpoints):
    d1 = np.linalg.norm(p - endpoints[(i+1) % 3])
    d2 = np.linalg.norm(p - endpoints[(i+2) % 3])
    total_distance = d1 + d2
    if total_distance > max_distance:
      max_distance = total_distance
      farthest_pt = p
  
  return farthest_pt


def skeletonizeContour(self, im0, currContour, animalNum, frameNum, returnSkeletonNoMatterWhat=0):
  
  if len(currContour.astype('int32')) <= 1:
    return []
  
  # Create mask from contour
  img = np.zeros((len(im0), len(im0[0])), dtype=np.uint8)
  currContour = currContour.astype('int32')
  cv2.drawContours(img, [currContour], -1, 255, thickness=cv2.FILLED)
  
  # Skeletonize
  binary = img > 0
  skeleton = skeletonize(binary)
  skeleton_img = skeleton.astype(np.uint8) * 255
  
  # Get skeleton points
  skeleton_points = np.column_stack(np.where(skeleton))
  skeleton_points = skeleton_points[:, ::-1]  # Swap columns to get (x,y) format
  
  if len(skeleton_points) == 0:
    return np.array([], dtype=np.int32).reshape(-1, 1, 2)
  
  # Find endpoints (points with only one neighbor)
  endpoints = []
  for point in skeleton_points:
    x, y = point
    neighbors = 0
    for i in range(-1, 2):
      for j in range(-1, 2):
        if i == 0 and j == 0:
          continue
        nx, ny = x + i, y + j
        if 0 <= nx < skeleton_img.shape[1] and 0 <= ny < skeleton_img.shape[0]:
          if skeleton_img[ny, nx] > 0:
            neighbors += 1
    if neighbors == 1:
      endpoints.append(point)
  
  if len(endpoints) > 3 and not(returnSkeletonNoMatterWhat):
    return []
  
  if len(endpoints) >= 2:
    
    if len(endpoints) == 3:
      start_point = farthest_point(endpoints)
      start_point = tuple(start_point)
    else:
      endpoints.sort(key=lambda p: p[0], reverse=True)
      start_point = tuple(endpoints[0])
    
    # Trace the path from head to tail
    ordered_points = []
    current = start_point
    visited = set()
    
    while True:
      ordered_points.append(current)
      visited.add(current)
        
      # Find neighbors
      x, y = current
      neighbors = []
      for i in range(-1, 2):
        for j in range(-1, 2):
          if i == 0 and j == 0:
            continue
          nx, ny = x + i, y + j
          if 0 <= nx < skeleton_img.shape[1] and 0 <= ny < skeleton_img.shape[0]:
            if skeleton_img[ny, nx] > 0 and (nx, ny) not in visited:
              neighbors.append((nx, ny))
      
      if not neighbors:
        break
          
      # Move to the next point
      current = neighbors[0]
    
    ordered_points = np.array(ordered_points, dtype=np.int32)
  else:
    # If we couldn't find clear endpoints, fall back to simply using the skeleton points
    print(len(endpoints), "Not found for animalNum:", animalNum, " and frameNum:", frameNum)
    ordered_points = skeleton_points
  
  # Downsample if needed
  nbTailPoints = int(self._hyperparameters.get("nbTailPoints", 10))
  if len(ordered_points) > nbTailPoints or len(ordered_points) < nbTailPoints:
    indices = np.linspace(0, len(ordered_points) - 1, nbTailPoints, dtype=int)
    ordered_points = ordered_points[indices]
  
  # Reshape to the required format
  ordered_points = ordered_points.reshape(-1, 1, 2)
  
  return ordered_points


import zebrazoom.dataAPI as dataAPI
import numpy as np
from scipy.signal import savgol_filter
from scipy.interpolate import interp1d

def resample_curve(points, num_points):
  deltas = np.diff(points, axis=0)
  dists = np.sqrt((deltas**2).sum(axis=1))
  cumulative = np.insert(np.cumsum(dists), 0, 0)

  if cumulative[-1] == 0:
    # All points are the same, just duplicate
    return np.tile(points[0], (num_points, 1))

  fx = interp1d(cumulative, points[:, 0], kind='linear')
  fy = interp1d(cumulative, points[:, 1], kind='linear')

  equal_spacing = np.linspace(0, cumulative[-1], num_points)
  x_new = fx(equal_spacing)
  y_new = fy(equal_spacing)
  return np.stack([x_new, y_new], axis=1)

def smooth_and_resample(dataHead, dataTailX, dataTailY, window_length=9, polyorder=2):
  # Smooth head
  head_x = savgol_filter(dataHead[:, 0], window_length, polyorder)
  head_y = savgol_filter(dataHead[:, 1], window_length, polyorder)
  dataHeadSmoothed = np.stack([head_x, head_y], axis=1)

  # Smooth tail
  tail_x_smooth = savgol_filter(dataTailX, window_length, polyorder, axis=0)
  tail_y_smooth = savgol_filter(dataTailY, window_length, polyorder, axis=0)

  # Resample tail to ensure equal spacing (per frame)
  N, T = tail_x_smooth.shape
  resampled_tail_x = np.zeros((N, T))
  resampled_tail_y = np.zeros((N, T))

  for i in range(N):
    tail_curve = np.stack([tail_x_smooth[i], tail_y_smooth[i]], axis=1)
    resampled = resample_curve(tail_curve, T)
    resampled_tail_x[i] = resampled[:, 0]
    resampled_tail_y[i] = resampled[:, 1]

  return dataHeadSmoothed, resampled_tail_x, resampled_tail_y


def smoothTailAngleOverTime(videoName: str, nbWells: int, nbAnimalsPerWell: int, freqAlgoPosFollow: int, startTimeInSeconds = None, endTimeInSeconds = None):
  
  for numWell in range(nbWells):
    
    for numAnimal in range(nbAnimalsPerWell):
      dataHead = dataAPI.getDataPerTimeInterval(videoName, numWell, numAnimal, startTimeInSeconds, endTimeInSeconds, "HeadPos")
      dataTailX = dataAPI.getDataPerTimeInterval(videoName, numWell, numAnimal, startTimeInSeconds, endTimeInSeconds, "TailPosX")
      dataTailY = dataAPI.getDataPerTimeInterval(videoName, numWell, numAnimal, startTimeInSeconds, endTimeInSeconds, "TailPosY")
      
      dataHead, dataTailX, dataTailY = smooth_and_resample(dataHead, dataTailX, dataTailY)
      
      dataAPI.setDataPerTimeInterval(videoName, numWell, numAnimal, startTimeInSeconds, endTimeInSeconds, "HeadPos", dataHead)
      dataAPI.setDataPerTimeInterval(videoName, numWell, numAnimal, startTimeInSeconds, endTimeInSeconds, "TailPosX", dataTailX)
      dataAPI.setDataPerTimeInterval(videoName, numWell, numAnimal, startTimeInSeconds, endTimeInSeconds, "TailPosY", dataTailY)



from typing import Tuple

from ._openResultsFile import openResultsFile


def getWellCoordinates(videoName: str, numWell: int) -> Tuple[int, int, int, int]:
  with openResultsFile(videoName, 'r') as results:
    return results["wellPositions/well" + str(numWell)].attrs["topLeftX"], results["wellPositions/well" + str(numWell)].attrs["topLeftY"], results["wellPositions/well" + str(numWell)].attrs["lengthX"], results["wellPositions/well" + str(numWell)].attrs["lengthY"]



import numpy as np
import cv2

def trackTailWithYOLO_getContours(self, results):
  curr_contours = []
  findContoursWithXY = False
  if findContoursWithXY:
    # Use the contours provided by ultralytics (only the biggest contour is provided)
    for idx in range(self._hyperparameters["nbAnimalsPerWell"]):
      curr_contours.append(results[0].masks.xy[idx] if len(results[0].masks.xy) > idx else np.array([[]]))
  else:
    # Finds the contour from the mask image provided by ultralytics (to take into account all the contours provided by ultralytics, not just the biggest one)
    totContours = []
    if results[0].masks is not None:
      for mask in results[0].masks.data:
        mask_np = (mask.cpu().numpy() * 255).astype(np.uint8)
        contours, _ = cv2.findContours(mask_np, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
        if len(contours) > 1:
          # If more than one contour was detected, need to merge all the contours using OpenCv's watershed
          dist = cv2.distanceTransform(mask_np, cv2.DIST_L2, 5)
          _, sure_fg = cv2.threshold(dist, 0.5 * dist.max(), 255, 0)
          sure_fg = np.uint8(sure_fg)
          # Sure background by dilating original mask
          kernel = np.ones((3,3), np.uint8)
          sure_bg = cv2.dilate(mask_np, kernel, iterations=2)
          # Unknown region is bg - fg
          unknown = cv2.subtract(sure_bg, sure_fg)
          # Marker labelling
          _, markers = cv2.connectedComponents(sure_fg)
          # Add 1 to all labels so that sure background is 1, unknown is 0
          markers = markers + 1
          markers[unknown == 255] = 0
          # Watershed needs a 3-channel image
          mask_color = cv2.cvtColor(mask_np, cv2.COLOR_GRAY2BGR)
          markers = cv2.watershed(mask_color, markers)
          # Result: areas where markers > 1 are segmented
          # Optional: merge all non-boundary regions into a single mask
          mask_np = np.uint8(markers > 1) * 255
          contours, _ = cv2.findContours(mask_np, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
        
        contours = np.vstack(contours)
        totContours.append(contours)
    
    curr_contours = [cnt.reshape(-1, 2).astype(np.float32) for cnt in totContours]
    scale_x = int(self._hyperparameters["videoWidth"]) / 640 #int(np.sum(len(results[0].masks.data[i]) for i in range(len(results[0].masks.data))))
    scale_y = int(self._hyperparameters["videoHeight"]) / 640 #int(len(results[0].masks.data[0][0]))
    curr_contours = [np.column_stack((contour[:, 0] * scale_x, contour[:, 1] * scale_x)) for contour in curr_contours]
  
  return curr_contours



from scipy.optimize import linear_sum_assignment
import zebrazoom.dataAPI as dataAPI
import matplotlib.pyplot as plt
from scipy import stats
import pandas as pd
import numpy as np
import shutil
import math
import h5py
import os

def findAndSmooth(data, threshold=3, k=10):
  n = len(data)
  if k >= n:
    raise ValueError("k must be smaller than the length of the data.")
  
  # Calculate distances between points n and n+k
  distances = np.zeros(n - k)
  for i in range(n - k):
    distances[i] = np.linalg.norm(data[i + k] - data[i])
  
  # Identify stationary segments
  stationary_mask = distances <= threshold
  result = data.copy()
  
  i = 0
  while i < n - k:
    if stationary_mask[i]:
      # Start of a stationary segment
      start_idx = i
      while i < n - k and stationary_mask[i]:
        i += 1
      # Replace the segment with the coordinates at the start
      result[start_idx + 1:i + k] = data[start_idx]
    else:
      i += 1
  
  return result


def findAndSmoothStationaryHeadPositions(videoName: str, nbWells: int, nbAnimalsPerWell: int, findAndSmoothStationaryDist: int = 3, findAndSmoothStationaryIndexDiff: int = 10):

  for numWell in range(nbWells):
    dataForEachAnimal = []
    for numAnimal in range(nbAnimalsPerWell):
      data = dataAPI.getDataPerTimeInterval(videoName, numWell, numAnimal, None, None, "HeadPos")
      smoothed_data = findAndSmooth(data, findAndSmoothStationaryDist, findAndSmoothStationaryIndexDiff)
      dataAPI.setDataPerTimeInterval(videoName, numWell, numAnimal, None, None, "HeadPos", smoothed_data)



import numpy as np
import math

def calculateHeading(coords):

  if len(coords) < 3:
      raise ValueError("At least three points are required.")

  coords = np.array(coords)

  p1, p2, p3 = coords[:3]

  v1 = p2 - p1
  v2 = p3 - p2
  
  v = (v1 + v2) / 2
  if v[0] != 0:
    angle = math.atan(v[1] / v[0])
  else:
    if v[1] > 0:
      angle = - math.pi / 2
    else:
      angle = math.pi / 2
  if v[0] > 0:
    angle = (angle + math.pi) % (2 * math.pi)

  return angle



from zebrazoom.code.tracking.customTrackingImplementations.fastFishTracking.trackTail import trackTail
import cv2

def trackTailWithClassicalCV(self, frame, frameGaussianBlurForHeadPosition, xmin, ymin, xmax, ymax, wellNum, animalNum, frameNum):
  
  # import pdb
  # pdb.set_trace()
  margin = 30
  xmin = xmin - margin if xmin - margin >= 0 else 0
  ymin = ymin - margin if ymin - margin >= 0 else 0
  xmax = xmax + margin if xmax + margin < len(frame[0]) else len(frame[0]) - 1
  ymax = ymax + margin if ymax + margin < len(frame) else len(frame) - 1
  
  frameROIGaussianBlurForHeadPosition = frameGaussianBlurForHeadPosition[int(ymin):int(ymax), int(xmin):int(xmax)].copy()
  frameROIGaussianBlurForHeadPosition = cv2.cvtColor(frameROIGaussianBlurForHeadPosition, cv2.COLOR_BGR2GRAY)
  
  frameROI = frame[int(ymin):int(ymax), int(xmin):int(xmax)].copy()
  frameROI = cv2.cvtColor(frameROI, cv2.COLOR_BGR2GRAY)
  
  # import zebrazoom.code.util as util
  # util.showFrame(frameROI, title="tailTracking")
  
  (minVal, maxVal, headPosition, maxLoc) = cv2.minMaxLoc(frameROIGaussianBlurForHeadPosition)
  
  self._trackingHeadTailAllAnimalsList[wellNum][animalNum, frameNum-self._firstFrame][0][0] = headPosition[0] + xmin - self._wellPositions[wellNum]['topLeftX']
  self._trackingHeadTailAllAnimalsList[wellNum][animalNum, frameNum-self._firstFrame][0][1] = headPosition[1] + ymin - self._wellPositions[wellNum]['topLeftY']
  
  # Tail Tracking
  if not("onlyRecenterHeadPosition" in self._hyperparameters) or (self._hyperparameters["onlyRecenterHeadPosition"] == 0):
    a, self._lastFirstTheta[wellNum] = trackTail(self, frameROI, headPosition, self._hyperparameters, wellNum, frameNum, self._lastFirstTheta[wellNum])
    if len(a):
      a[0, :, 0] += int(xmin) - self._wellPositions[wellNum]['topLeftX']
      a[0, :, 1] += int(ymin) - self._wellPositions[wellNum]['topLeftY']
      self._trackingHeadTailAllAnimalsList[wellNum][animalNum][frameNum-self._firstFrame][:len(a[0])] = a
    else:
      self._trackingHeadTailAllAnimalsList[wellNum][animalNum][frameNum-self._firstFrame] = self._trackingHeadTailAllAnimalsList[wellNum][animalNum][frameNum-self._firstFrame-1]
      # animalNotTracked[animalId] = 1


from zebrazoom.code.tracking.customTrackingImplementations.yolov11.trackTailWithYOLO_linkBetweenFrames import update_fixed_tracking
from zebrazoom.code.tracking.customTrackingImplementations.yolov11.trackTailWithYOLO_getContours import trackTailWithYOLO_getContours
from zebrazoom.code.tracking.customTrackingImplementations.yolov11.trackTailWithYOLO_skeletonizeContour import skeletonizeContour
from zebrazoom.code.tracking.customTrackingImplementations.yolov11.trackTailWithYOLO_invertSkeletonIfNecessary import invertSkeletonIfNecessaryUsingThePast, invertSkeletonIfNecessaryUsingTheDarkEyes
from zebrazoom.code.tracking.customTrackingImplementations.yolov11.calculateHeading import calculateHeading
import numpy as np
import cv2


def moving_average_smoothing(points, window_size=3):
    
  # Ensure that the window size is odd to have a center element
  window_size = window_size if window_size % 2 != 0 else window_size + 1
  half_window = window_size // 2
  smoothed_points = []

  # Apply moving average smoothing
  for i in range(len(points)):
    start = max(0, i - half_window)
    end = min(len(points), i + half_window + 1)
    window = points[start:end]
    smoothed_points.append(np.mean(window, axis=0))
  
  return np.array(smoothed_points)


def trackTailWithYOLO(self, im0, results, frameNum, wellNum, prev_contours, disappeared_counts, numAlreadyInvertedWithThePast):
  
  if frameNum == max(0, self._firstFrame):
    # Getting contours for first frame to track
    curr_contours = trackTailWithYOLO_getContours(self, results)
  else:
    # For subsequent frames, getting contours and linking them to the corresponding previous contours found
    curr_contours, disappeared_counts = update_fixed_tracking(self, results, prev_contours, disappeared_counts, self._hyperparameters["nbAnimalsPerWell"])
    prev_contours = curr_contours
  
  if im0.ndim == 3:
    im0b = cv2.cvtColor(im0, cv2.COLOR_BGR2GRAY)
  else:
    im0b = im0
  
  smooth_factor_max = 20
  
  currContourOri = []
  
  for idxContour, currContour in enumerate(curr_contours):
    # Iteratively increasing smoothing of the contour until good skeleton is found
    currContourOri = currContour.copy()
    smooth_factor = 3
    currContour = moving_average_smoothing(currContourOri, smooth_factor)
    skeleton_points = skeletonizeContour(self, im0, currContour, idxContour, frameNum)
    while smooth_factor < smooth_factor_max and len(skeleton_points) == 0:
      smooth_factor += 2
      currContour = moving_average_smoothing(currContourOri, smooth_factor)
      skeleton_points = skeletonizeContour(self, im0, currContour, idxContour, frameNum)
    if len(skeleton_points) == 0:
      skeleton_points = skeletonizeContour(self, im0, currContour, idxContour, frameNum, 1)
    # Inverting skeleton point when necessary
    skeleton_points = invertSkeletonIfNecessaryUsingTheDarkEyes(im0b, currContour, skeleton_points)
    if frameNum-self._firstFrame-1 >= 0:
      skeleton_points = invertSkeletonIfNecessaryUsingThePast(self, skeleton_points, wellNum, frameNum, numAlreadyInvertedWithThePast, idxContour, self._hyperparameters.get("maxConsecutiveFramesInversion", 25))
    # Storing skeleton points into the output
    if len(skeleton_points):
      skeleton_points[:, 0, :][0] = 2 * skeleton_points[:, 0, :][1] - skeleton_points[:, 0, :][2] # Need to improve this in the future!
      self._trackingHeadTailAllAnimalsList[wellNum][idxContour][frameNum-self._firstFrame][:len(skeleton_points)] = skeleton_points[:, 0, :]
      self._trackingHeadingAllAnimalsList[wellNum][idxContour][frameNum-self._firstFrame] = calculateHeading(skeleton_points[:, 0, :])

  
  # Debugging visualization option
  debug1 = self._hyperparameters["debugTracking"]
  colorsForMask = [(0, 255, 0), (255, 0, 0), (0, 0, 255), (255, 255, 0), (255, 0, 255), (0, 255, 255)] # need to add more colors for debugging if tracking more than 6 animals
  if debug1:
    annotator = Annotator(im0, line_width=2)
    for idxContour, currContour in enumerate(curr_contours):
      if len(currContour):
        annotator.seg_bbox(mask=currContour, mask_color=colorsForMask[idxContour], txt_color=annotator.get_txt_color(colorsForMask[idxContour]))
      pts = np.array(self._trackingHeadTailAllAnimalsList[wellNum][idxContour][frameNum - self._firstFrame], dtype=np.int32)
      pts = pts.reshape((-1, 1, 2))  # Required shape for cv2.polylines
      cv2.polylines(im0, [pts], isClosed=False, color=(0, 0, 255), thickness=1)
    import zebrazoom.code.util as util
    util.showFrame(im0, title="write title here")
  
  prev_contours = curr_contours 

  return [prev_contours, disappeared_counts]



# from zebrazoom.code.tracking.customTrackingImplementations.fastFishTracking.detectMovementWithRawVideoInsideTracking import detectMovementWithRawVideoInsideTracking
from zebrazoom.code.tracking.customTrackingImplementations.yolov11.trackTailWithClassicalCV import trackTailWithClassicalCV
from zebrazoom.code.tracking.customTrackingImplementations.yolov11.trackTailWithYOLO import trackTailWithYOLO
from zebrazoom.code.extractParameters import extractParameters

import cv2
import os
import zebrazoom.videoFormatConversion.zzVideoReading as zzVideoReading
import numpy as np
from zebrazoom.code.tracking import register_tracking_method
from ..._fasterMultiprocessingBase import BaseFasterMultiprocessing

import zebrazoom.code.tracking

from ultralytics import YOLO


class Yolov11basedTracking(BaseFasterMultiprocessing):
  def __init__(self, videoPath, wellPositions, hyperparameters):
    super().__init__(videoPath, wellPositions, hyperparameters)
    if self._hyperparameters["eyeTracking"]:
      self._trackingEyesAllAnimalsList = [np.zeros((self._hyperparameters["nbAnimalsPerWell"], self._lastFrame-self._firstFrame+1, 8))
                                          for _ in range(self._hyperparameters["nbWells"])]
    else:
      self._trackingEyesAllAnimals = 0

    # if not(self._hyperparameters["nbAnimalsPerWell"] > 1) and not(self._hyperparameters["headEmbeded"]) and (self._hyperparameters["findHeadPositionByUserInput"] == 0) and (self._hyperparameters["takeTheHeadClosestToTheCenter"] == 0):
    self._trackingProbabilityOfGoodDetectionList = [np.zeros((self._hyperparameters["nbAnimalsPerWell"], self._lastFrame-self._firstFrame+1))
                                                      for _ in range(self._hyperparameters["nbWells"])]
    # else:
      # self._trackingProbabilityOfGoodDetectionList = 0
    
    if self._hyperparameters["headingCalculationMethod"]:
      self._trackingProbabilityOfHeadingGoodCalculation = [np.zeros((self._hyperparameters["nbAnimalsPerWell"], self._lastFrame-self._firstFrame+1)) for _ in range(self._hyperparameters["nbWells"])]
    else:
      self._trackingProbabilityOfHeadingGoodCalculation = 0

  def _adjustParameters(self, i, frame, widgets):
    return None

  def _formatOutput(self):
    if self._hyperparameters["postProcessMultipleTrajectories"]:
      for wellNumber in range(self._firstWell, self._lastWell + 1):
        self._postProcessMultipleTrajectories(self._trackingHeadTailAllAnimalsList[wellNumber], self._trackingProbabilityOfGoodDetectionList[wellNumber])
    
    if "postProcessHeadingWithTrajectory_minDist" in self._hyperparameters and self._hyperparameters["postProcessHeadingWithTrajectory_minDist"]:
      for wellNumber in range(self._firstWell, self._lastWell + 1):
        self._postProcessHeadingWithTrajectory(self._trackingHeadTailAllAnimalsList[wellNumber], self._trackingHeadingAllAnimalsList[wellNumber], self._trackingProbabilityOfHeadingGoodCalculation[wellNumber])
    
    return {wellNumber: extractParameters([self._trackingHeadTailAllAnimalsList[wellNumber], self._trackingHeadingAllAnimalsList[wellNumber], [], 0, 0] + ([self._auDessusPerAnimalIdList[wellNumber]] if self._auDessusPerAnimalIdList is not None else []), wellNumber, self._hyperparameters, self._videoPath, self._wellPositions, self._background, 0, self._trackingProbabilityOfGoodDetectionList[wellNumber])
            for wellNumber in range(self._firstWell, self._lastWell + 1)}

  def run(self):
    
    if os.path.exists(self._hyperparameters["DLmodelPath"]):
      model = YOLO(self._hyperparameters["DLmodelPath"])
    elif os.path.exists(os.path.join('zebrazoom', 'configuration', self._hyperparameters["DLmodelPath"])):
      model = YOLO(os.path.join('zebrazoom', 'configuration', self._hyperparameters["DLmodelPath"]))
    elif os.path.exists(os.path.join('ZebraZoom', 'zebrazoom', 'configuration', self._hyperparameters["DLmodelPath"])):
      model = YOLO(os.path.join('ZebraZoom', 'zebrazoom', 'configuration', self._hyperparameters["DLmodelPath"]))
    else:
      print("Path to DL model not found")
      raise ValueError("Path to DL model not found")
    
    if self._hyperparameters["trackTail"] or (("onlyRecenterHeadPosition" in self._hyperparameters) and (self._hyperparameters["onlyRecenterHeadPosition"] == 1)):
      self._lastFirstTheta = np.zeros(len(self._wellPositions))
      self._lastFirstTheta[:] = -99999
      
    if "trackTailWithYOLO" in self._hyperparameters and self._hyperparameters["trackTailWithYOLO"]:
      prev_contours = [0] * self._hyperparameters["nbAnimalsPerWell"]
      disappeared_counts = [0] * self._hyperparameters["nbAnimalsPerWell"]
      numAlreadyInvertedWithThePast = [0] * int(self._hyperparameters["nbAnimalsPerWell"])
    
    wellNum = 0
    
    # Open the video
    cap = cv2.VideoCapture(self._videoPath)
    frameNum = 0
    while cap.isOpened() and frameNum <= self._lastFrame:
      
      ret, frame = cap.read()
      if not ret:
        break
        
      # if self._hyperparameters["detectMovementWithRawVideoInsideTracking"]:
        # prevFrame = detectMovementWithRawVideoInsideTracking(self, k, frame)
        # if ("detectMovementCompareWithTheFuture" in self._hyperparameters) and self._hyperparameters["detectMovementCompareWithTheFuture"]:
          # frame = prevFrame
      
      if self._hyperparameters["trackTail"] or (("onlyRecenterHeadPosition" in self._hyperparameters) and (self._hyperparameters["onlyRecenterHeadPosition"] == 1)):
        frameGaussianBlur = frame.copy()
        paramGaussianBlur = self._hyperparameters["paramGaussianBlur"]
        frameGaussianBlur = cv2.GaussianBlur(frameGaussianBlur, (paramGaussianBlur, paramGaussianBlur), 0)
        if ("paramGaussianBlurForHeadPosition" in self._hyperparameters) and self._hyperparameters["paramGaussianBlurForHeadPosition"]:
          frameGaussianBlurForHeadPosition = frame.copy()
          frameGaussianBlurForHeadPosition = cv2.GaussianBlur(frameGaussianBlurForHeadPosition, (self._hyperparameters["paramGaussianBlurForHeadPosition"], self._hyperparameters["paramGaussianBlurForHeadPosition"]), 0)
        else:
          frameGaussianBlurForHeadPosition = frameGaussianBlur
      
      if frameNum >= self._firstFrame:
      
        if frameNum % self._hyperparameters["freqAlgoPosFollow"] == 0:
          print("YOLO tracking at frame", frameNum)
        
        if "yolo11MinConf" in self._hyperparameters:
          results = model(frame, verbose=False, conf=float(self._hyperparameters["yolo11MinConf"]))
        else:
          results = model(frame, verbose=False)
          
        # Set to True for debugging
        if False:
          result = results[0]
          # Draw contours if masks are available
          if result.masks is not None:
            for idx, mask in enumerate(result.masks.data):
              # Convert mask tensor to uint8 numpy array
              mask_np = (mask.cpu().numpy() * 255).astype(np.uint8)
              # Find contours
              contours, _ = cv2.findContours(mask_np, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
              # Draw contours on the original frame
              if idx == 0:
                cv2.drawContours(frame, contours, -1, (0, 255, 0), 2)
              else:
                cv2.drawContours(frame, contours, -1, (0, 0, 255), 2)
          # Show the result
          import zebrazoom.code.util as util
          util.showFrame(frame, title="write title here")
          ################          
        
        if "trackTailWithYOLO" in self._hyperparameters and self._hyperparameters["trackTailWithYOLO"]:
          
          [prev_contours, disappeared_counts] = trackTailWithYOLO(self, frame, results, frameNum, wellNum, prev_contours, disappeared_counts, numAlreadyInvertedWithThePast)
          
        else:
          animalNum = 0

          for animalNum, result in enumerate(results[0].boxes):
            
            if animalNum < self._hyperparameters["nbAnimalsPerWell"]:
              xmin, ymin, xmax, ymax = result.xyxy[0]
              
              if self._hyperparameters["trackTail"] or (("onlyRecenterHeadPosition" in self._hyperparameters) and (self._hyperparameters["onlyRecenterHeadPosition"] == 1)):
                trackTailWithClassicalCV(self, frameGaussianBlur, frameGaussianBlurForHeadPosition, xmin, ymin, xmax, ymax, wellNum, animalNum, frameNum)
              else:
                xCenter = float((xmin + xmax) / 2)
                yCenter = float((ymin + ymax) / 2)
                self._trackingHeadTailAllAnimalsList[wellNum][animalNum, frameNum-self._firstFrame][0][0] = int(xCenter)
                self._trackingHeadTailAllAnimalsList[wellNum][animalNum, frameNum-self._firstFrame][0][1] = int(yCenter)
              
              self._trackingProbabilityOfGoodDetectionList[wellNum][animalNum, frameNum-self._firstFrame] = float(result.conf[0])
      
      frameNum += 1
    
    if frameNum-self._firstFrame < len(self._trackingHeadTailAllAnimalsList[wellNum][0]):
      for animalNum in range(self._hyperparameters["nbAnimalsPerWell"]):
        self._trackingHeadTailAllAnimalsList[wellNum][animalNum, frameNum-self._firstFrame][0][0] = self._trackingHeadTailAllAnimalsList[wellNum][animalNum, frameNum-self._firstFrame-1][0][0]
        self._trackingHeadTailAllAnimalsList[wellNum][animalNum, frameNum-self._firstFrame][0][1] = self._trackingHeadTailAllAnimalsList[wellNum][animalNum, frameNum-self._firstFrame-1][0][1]
    
    cap.release()
    
    return self._formatOutput()


zebrazoom.code.tracking.register_tracking_method('yolov11basedTracking', Yolov11basedTracking)



from zebrazoom.code.extractParameters import extractParameters

import cv2
import zebrazoom.videoFormatConversion.zzVideoReading as zzVideoReading
import numpy as np
import math
from zebrazoom.code.tracking import register_tracking_method
from ..._eyeTracking import EyeTrackingMixin
from ..._fasterMultiprocessingBase import BaseFasterMultiprocessing
from ..._getImages import GetImagesMixin

from zebrazoom.code.tracking.customTrackingImplementations.multipleCenterOfMassTracking._headTrackingHeadingCalculation import _headTrackingHeadingCalculation

import zebrazoom.code.tracking
from collections import deque


class MultipleCenterOfMassTracking(BaseFasterMultiprocessing, EyeTrackingMixin, GetImagesMixin):
  def __init__(self, videoPath, wellPositions, hyperparameters):
    super().__init__(videoPath, wellPositions, hyperparameters)

    self._trackingProbabilityOfGoodDetectionList = [np.zeros((self._hyperparameters["nbAnimalsPerWell"], self._lastFrame-self._firstFrame+1)) for _ in range(self._hyperparameters["nbWells"])]
    
    if self._hyperparameters["headingCalculationMethod"]:
      self._trackingProbabilityOfHeadingGoodCalculation = [np.zeros((self._hyperparameters["nbAnimalsPerWell"], self._lastFrame-self._firstFrame+1)) for _ in range(self._hyperparameters["nbWells"])]
    else:
      self._trackingProbabilityOfHeadingGoodCalculation = 0

  def _adjustParameters(self, i, frame, unprocessedFrame, widgets):
    return None

  def _formatOutput(self):
    if self._hyperparameters["postProcessMultipleTrajectories"]:
      for wellNumber in range(self._firstWell, self._lastWell + 1):
        self._postProcessMultipleTrajectories(self._trackingHeadTailAllAnimalsList[wellNumber], self._trackingProbabilityOfGoodDetectionList[wellNumber])
    
    if "postProcessHeadingWithTrajectory_minDist" in self._hyperparameters and self._hyperparameters["postProcessHeadingWithTrajectory_minDist"]:
      for wellNumber in range(self._firstWell, self._lastWell + 1):
        self._postProcessHeadingWithTrajectory(self._trackingHeadTailAllAnimalsList[wellNumber], self._trackingHeadingAllAnimalsList[wellNumber], self._trackingProbabilityOfHeadingGoodCalculation[wellNumber])
    
    return {wellNumber: extractParameters([self._trackingHeadTailAllAnimalsList[wellNumber], self._trackingHeadingAllAnimalsList[wellNumber], [], 0, 0] + ([self._auDessusPerAnimalIdList[wellNumber]] if self._auDessusPerAnimalIdList is not None else []), wellNumber, self._hyperparameters, self._videoPath, self._wellPositions, self._background)
            for wellNumber in range(self._firstWell, self._lastWell + 1)}

  def run(self):
    self._background = self.getBackground()
    
    cap = zzVideoReading.VideoCapture(self._videoPath)
    if (cap.isOpened()== False):
      print("Error opening video stream or file")

    if "backgroundSubtractorKNN_history" in self._hyperparameters and self._hyperparameters["backgroundSubtractorKNN_history"]:
      if "backgroundSubtractorKNN_dist2Threshold" in self._hyperparameters and self._hyperparameters["backgroundSubtractorKNN_dist2Threshold"]:
        dist2Threshold = int(self._hyperparameters["backgroundSubtractorKNN_dist2Threshold"])
      else:
        dist2Threshold = 400
      fgbg = cv2.createBackgroundSubtractorKNN(int(self._hyperparameters["backgroundSubtractorKNN_history"]), dist2Threshold)
    else:
      fgbg = cv2.createBackgroundSubtractorKNN()
    
    for i in range(0, min(self._lastFrame - 1, 500), int(min(self._lastFrame - 1, 500) / 10)):
      cap.set(1, min(self._lastFrame - 1, 500) - i)
      ret, frame = cap.read()
      fgmask = fgbg.apply(frame)
    cap.release()
    cap = zzVideoReading.VideoCapture(self._videoPath)
    
    queue1 = deque()
    queue2 = deque()
    queue3 = deque()
    
    i = self._firstFrame

    if self._firstFrame:
      cap.set(1, self._firstFrame)

    previousFrames = None
    widgets = None
    while (i < self._lastFrame + 1):

      if (self._hyperparameters["freqAlgoPosFollow"] != 0) and (i % self._hyperparameters["freqAlgoPosFollow"] == 0):
        print("Tracking: frame:",i)
        if self._hyperparameters["popUpAlgoFollow"]:
          from zebrazoom.code.popUpAlgoFollow import prepend

      if self._hyperparameters["debugTracking"]:
        print("frame:",i)

      ret, frame = cap.read()

      if ret:
        
        frameOri = frame.copy()
        if self._hyperparameters["detectMovementWithRawVideoInsideTracking"]:
          frameOri = cv2.cvtColor(frameOri, cv2.COLOR_BGR2GRAY)
        
        if "removeShades" in self._hyperparameters and self._hyperparameters["removeShades"]:
          if len(queue1) >= 20:
            oldFrame1 = queue1.popleft()
          else:
            oldFrame1 = frameOri.copy()
          queue1.append(frameOri.copy())
          if len(queue2) >= 50:
            oldFrame2 = queue2.popleft()
          else:
            oldFrame2 = frameOri.copy()
          queue2.append(frameOri.copy())
          if len(queue3) >= 100:
            oldFrame3 = queue3.popleft()
          else:
            oldFrame3 = frameOri.copy()
          queue3.append(frameOri.copy())
        else:
          oldFrame1 = 0
          oldFrame2 = 0
          oldFrame3 = 0
        
        frame = fgbg.apply(frame)
        frame = 255 - frame

        for wellNumber in range(self._firstWell, self._lastWell + 1):
          minPixelDiffForBackExtract = self._hyperparameters["minPixelDiffForBackExtract"]
          xtop = self._wellPositions[wellNumber]['topLeftX']
          ytop = self._wellPositions[wellNumber]['topLeftY']
          lenX = self._wellPositions[wellNumber]['lengthX']
          lenY = self._wellPositions[wellNumber]['lengthY']
          grey = frame
          
          curFrame = grey[ytop:ytop+lenY, xtop:xtop+lenX].copy()
            
          if self._hyperparameters["paramGaussianBlur"]:
            blur = cv2.GaussianBlur(curFrame, (self._hyperparameters["paramGaussianBlur"], self._hyperparameters["paramGaussianBlur"]),0)
          else:
            blur = curFrame
          if "headingCalculationMethod" in self._hyperparameters:
            t, thresh1 = cv2.threshold(curFrame, 254, 255, cv2.THRESH_BINARY)
            t, thresh2 = cv2.threshold(curFrame, 254, 255, cv2.THRESH_BINARY)
          else:
            thresh1 = 0
            thresh2 = 0
          gray    = 0
          blurCopy = blur.copy()

          headPositionFirstFrame = 0

          # Head tracking and heading calculation
          lastFirstTheta = _headTrackingHeadingCalculation(self, i, blur, thresh1, thresh2, frameOri, self._hyperparameters["erodeSize"], int(cap.get(3)), int(cap.get(4)), self._trackingHeadingAllAnimalsList[wellNumber], self._trackingHeadTailAllAnimalsList[wellNumber], self._trackingProbabilityOfGoodDetectionList[wellNumber], headPositionFirstFrame, self._wellPositions[wellNumber]["lengthX"], 0, 0, wellNumber, [oldFrame1, oldFrame2, oldFrame3])

          self._debugTracking(i, self._trackingHeadTailAllAnimalsList[wellNumber], self._trackingHeadingAllAnimalsList[wellNumber], curFrame)

          if self._hyperparameters["freqAlgoPosFollow"]:
            if i % self._hyperparameters["freqAlgoPosFollow"] == 0:
              print("Tracking at frame", i)

        if self._hyperparameters["detectMovementWithRawVideoInsideTracking"]:
          previousFrames = self._detectMovementWithRawVideoInsideTracking(i, frameOri, previousFrames)
      
      paramsAdjusted = self._adjustParameters(i, blurCopy, frameOri, widgets)
      if paramsAdjusted is not None:
        i, widgets = paramsAdjusted
        if self._hyperparameters.get('backgroundSubtractorKNN_history'):
          fgbg.setHistory(self._hyperparameters['backgroundSubtractorKNN_history'])
        if self._hyperparameters.get('backgroundSubtractorKNN_dist2Threshold'):
          fgbg.setDist2Threshold(self._hyperparameters['backgroundSubtractorKNN_dist2Threshold'])
        for arr in self._trackingHeadTailAllAnimalsList:
          arr[:] = 0
        cap.set(1, i)
      else:
        i = i + 1
    
    if self._hyperparameters["detectMovementWithRawVideoInsideTracking"]:
      frameGapComparision = self._hyperparameters["frameGapComparision"]
      nbFrames = len(self._trackingHeadTailAllAnimalsList[0][0])
      for wellNumber in range(len(self._trackingHeadTailAllAnimalsList)):
        for animalId in range(len(self._trackingHeadTailAllAnimalsList[wellNumber])):
          self._auDessusPerAnimalIdList[wellNumber][animalId][:nbFrames-frameGapComparision] = self._auDessusPerAnimalIdList[wellNumber][animalId][frameGapComparision:nbFrames]
      
    cap.release()
    return self._formatOutput()


zebrazoom.code.tracking.register_tracking_method('multipleCenterOfMassTracking', MultipleCenterOfMassTracking)



from zebrazoom.code.tracking.customTrackingImplementations.yolov11.trackTailWithYOLO_getContours import trackTailWithYOLO_getContours
from scipy.optimize import linear_sum_assignment
from scipy.spatial import distance
import numpy as np

def track_fixed_objects(self, results, prev_contours, disappeared_counts, num_objects, 
                        max_disappeared=5, conf_threshold=0.25, max_distance=10):
    """
    Track a fixed number of objects using the Hungarian algorithm.
    
    Args:
        results: Detection results containing masks and confidence scores
        prev_contours: List of previous contours for each tracked object
        disappeared_counts: List tracking how many frames each object has been missing
        num_objects: Fixed number of objects to track
        max_disappeared: Maximum number of frames an object can disappear before being dropped
        conf_threshold: Confidence threshold for detection
        max_distance: Maximum distance for immediate association without considering disappeared count
    
    Returns:
        curr_contours: Updated contours for each tracked object
        disappeared_counts: Updated disappeared counts
    """
    
    masks = results[0].masks.xy
    
    # First-time initialization
    if not prev_contours:
        # Initialize tracking with up to num_objects detected objects
        if len(masks) > 0:
            # Get masks with confidence above threshold
            valid_mask_indices = [
                i for i, conf in enumerate(results[0].boxes.conf) 
                if conf > conf_threshold and i < len(masks)
            ]
            
            # Sort by confidence (highest first)
            valid_mask_indices.sort(key=lambda i: results[0].boxes.conf[i], reverse=True)
            
            # Take up to num_objects masks
            selected_indices = valid_mask_indices[:num_objects]
            
            # Create initial contours and disappeared counts
            curr_contours = trackTailWithYOLO_getContours(self, results)
            
            curr_disappeared_counts = [0] * len(curr_contours)
            
            # Pad with empty contours if we don't have enough detections
            if len(curr_contours) < num_objects:
                # Use placeholder contours for missing objects
                # A placeholder could be a small box in the center or corner of the frame
                placeholder = np.array([[0, 0], [0, 1], [1, 1], [1, 0]])
                
                for _ in range(num_objects - len(curr_contours)):
                    curr_contours.append(placeholder)
                    curr_disappeared_counts.append(max_disappeared)  # Mark as disappeared
            
            return curr_contours, curr_disappeared_counts
        else:
            # No detections, create placeholder contours
            placeholder = np.array([[0, 0], [0, 1], [1, 1], [1, 0]])
            curr_contours = [placeholder] * num_objects
            curr_disappeared_counts = [max_disappeared] * num_objects
            return curr_contours, curr_disappeared_counts
    
    # Ensure we have the right number of objects being tracked
    if len(prev_contours) != num_objects:
        # Adjust the number of contours to match num_objects
        if len(prev_contours) < num_objects:
            # Add placeholder contours
            placeholder = np.array([[0, 0], [0, 1], [1, 1], [1, 0]])
            prev_contours = prev_contours + [placeholder] * (num_objects - len(prev_contours))
        else:
            # Keep only the first num_objects contours
            prev_contours = prev_contours[:num_objects]
    
    # Ensure disappeared_counts matches prev_contours length
    if len(disappeared_counts) != len(prev_contours):
        disappeared_counts = [0] * len(prev_contours)
    
    # Initialize current contours with previous ones (in case no matches are found)
    curr_contours = prev_contours.copy()
    
    if len(masks) > 0:
        # Filter masks by confidence threshold
        valid_mask_indices = [
            i for i, conf in enumerate(results[0].boxes.conf) 
            if conf > conf_threshold and i < len(masks)
        ]
        
        if valid_mask_indices:
            valid_masks = trackTailWithYOLO_getContours(self, results)
            
            # Calculate cost matrix
            cost_matrix = np.zeros((num_objects, len(valid_masks)))
            
            for i, prev_contour in enumerate(prev_contours):
                for j, mask in enumerate(valid_masks):
                    if len(mask) > 1 and len(prev_contour) > 1:
                        d = distance.cdist(prev_contour, mask)
                        # Symmetric average of minimum distances
                        cost_matrix[i, j] = (d.min(axis=1).mean() + d.min(axis=0).mean()) / 2
                    else:
                        cost_matrix[i, j] = 1000000000000000000000000000000000000 # perhaps improve this later?
            
            # Apply Hungarian algorithm to find optimal assignment
            row_indices, col_indices = linear_sum_assignment(cost_matrix)
            
            # Reset all disappeared counters to increment them later if not matched
            temp_disappeared = [disappeared_count + 1 for disappeared_count in disappeared_counts]
            # Update contours and reset disappeared counters for matched objects
            for row_idx, col_idx in zip(row_indices, col_indices):
                # Only consider matches where the distance is acceptable
                if cost_matrix[row_idx, col_idx] < max_distance or disappeared_counts[row_idx] >= max_disappeared // 2:
                    curr_contours[row_idx] = valid_masks[col_idx]
                    temp_disappeared[row_idx] = 0
            
            disappeared_counts = temp_disappeared
    else:
        # No detections, increment all disappeared counters
        disappeared_counts = [count + 1 for count in disappeared_counts]
    # Replace contours that have been disappeared too long with the best unmatched detections
    if len(masks) > 0:
        # Find contours that have been disappeared for too long
        long_disappeared_indices = [i for i, count in enumerate(disappeared_counts) 
                                   if count >= max_disappeared]
        
        if long_disappeared_indices:
            # Get unmatched detections with confidence above threshold
            matched_mask_indices = set([col_idx for _, col_idx in zip(row_indices, col_indices)])
            unmatched_mask_indices = [i for i in range(len(valid_masks)) 
                                      if i not in matched_mask_indices]
            
            # Sort by confidence (highest first)
            unmatched_mask_indices.sort(
                key=lambda i: valid_mask_indices[i] if i < len(valid_mask_indices) else 0, 
                reverse=True
            )
            
            # Replace disappeared contours with new detections
            for i, disappeared_idx in enumerate(long_disappeared_indices):
                if i < len(unmatched_mask_indices):
                    curr_contours[disappeared_idx] = valid_masks[unmatched_mask_indices[i]]
                    disappeared_counts[disappeared_idx] = 0
    
    return curr_contours, disappeared_counts


def update_fixed_tracking(self, results, prev_contours, disappeared_counts, num_objects, max_disappeared=5):
    """
    Main tracking function that maintains a fixed number of tracked objects.
    
    Args:
        results: Detection results containing masks and confidence scores
        prev_contours: List of previous contours for each tracked object
        disappeared_counts: List tracking how many frames each object has been missing
        num_objects: Fixed number of objects to track
        max_disappeared: Maximum number of frames an object can disappear
    
    Returns:
        updated_contours: Updated contours after tracking step (always num_objects length)
        updated_disappeared_counts: Updated disappeared counts (always num_objects length)
    """
    return track_fixed_objects(
        self, results, prev_contours, disappeared_counts, num_objects, max_disappeared
    )

